/*
 * This file is part of W3Olista, the statistics gathering program for
 * the WWW.                  This sourcecode is (c) Frank Pilhofer '95
 *
 * It is  distributed without warranty. The complete license terms for
 * this software  are included in  the file 'License.html'  in the doc
 * subdirectory. If  your copy came  without this file, you should get
 * it on the Web at    http://www.uni-frankfurt.de/~fp/Tools/W3Olista/
 * or contact the author,               fp@informatik.uni-frankfurt.de
 */


#ifdef RESOLVEADDR

/*
 * Look up domains
 * This once was done with a hashtable, but this changed after designing
 * the host lookup file. To add to the file, we need a sorted listing of
 * ip numbers, which a hash isn't. But maintaining the linear table is
 * not really that much more expensive - the DNS lookups still need most
 * of the time.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include "toolbox.h"
#include "hosts.h"

extern int h_errno;

/*
 * linux doesn't define this one
 */

#ifndef AF_INET
#define AF_INET		2
#endif

/*
 * size of table
 */

#ifndef LOOKUPSIZE
#define LOOKUPSIZE	256
#endif

/*
 * table entry
 */

static struct lookuptable_ {
  int used, tstamp;
  unsigned long address;
  char *domnam;
} *lookuptable = NULL;

/*
 * array for properly sorting the IP addresses
 */

static int tableused  = 0;
static int tabletime  = 0;

static int havehffile = 0;	/* have we openend a hosts file         */
static int logipaddrs = 0;	/* shall we save unresolvable addresses */
static int hfreadonly = 0;	/* shall we use the hosts file readonly */
static int hardlookup = 0;	/* shall we really DNS lookup IPs       */

#ifdef DEBUG2
static int totlookup  = 0;	/* number of addresses we've looked up  */
static int totintable = 0;	/* number of addresses found in table   */
static int totinhfile = 0;	/* number of addresses found in hfile   */
static int totdnslu   = 0;	/* number of addresses looked up        */
static int totdnssucc = 0;	/* number of succeeded dns lookups      */
#endif

void
InitLookup (char *hfile, int ro, int logip, int dolookup)
{
  if ((lookuptable = 
       (struct lookuptable_ *) 
       malloc (LOOKUPSIZE * sizeof (struct lookuptable_))) == NULL)
    return;

  memset (lookuptable, '\0', LOOKUPSIZE * sizeof (struct lookuptable_));

  tableused  = 0;
  tabletime  = (int) time (NULL);
  hardlookup = dolookup;

  if (hfile) {
    if (HTOpenFile (hfile) == 0) {
      havehffile = 1;
      logipaddrs = logip;
      hfreadonly = ro;
    }
    else
      havehffile = 0;
  }
  else
    havehffile = 0;

#ifdef DEBUG2
  totlookup  = 0;
  totintable = 0;
  totinhfile = 0;
  totdnslu   = 0;
  totdnssucc = 0;
#endif
}

/*
 * callback for adding to the hosts file
 */

static int
AddDataCallback (int ctr, unsigned long *aptr, char *buffer, int *count)
{
  static int index;

  if (ctr == 0)
    index = 0;

  while (index<tableused &&
	 (lookuptable[index].used == 0 ||
	  (lookuptable[index].domnam == NULL && !logipaddrs)))
    index++;

  if (index >= tableused)
    return 1;

  *aptr = lookuptable[index].address;

  if (lookuptable[index].domnam == NULL)
    sprintf (buffer, "%lu.%lu.%lu.%lu",
	     (lookuptable[index].address >> 24) & 255,
	     (lookuptable[index].address >> 16) & 255,
	     (lookuptable[index].address >>  8) & 255,
	     (lookuptable[index].address      ) & 255);
  else
    strcpy (buffer, lookuptable[index].domnam);

  *count = lookuptable[index].used;
  index++;

  return 0;
}

void
DestroyLookup (void)
{
  int index;

#ifdef DEBUG2
  fprintf (stderr, "\n\n\
Lookup Statistics\n\
-----------------\n\
Total Calls to Lookup        : %d\n\
Number of IPs found in table : %d\n\
Number of IPs found in file  : %d\n\
Number of DNS lookups done   : %d\n\
Number of successful lookups : %d\n\
\n\
",
	   totlookup,
	   totintable,
	   totinhfile,
	   totdnslu,
	   totdnssucc);
#endif

  /*
   * if we have a hosts file not opened for readonly, we add to it
   */

  if (havehffile && !hfreadonly && lookuptable)
    HTAddEntries (AddDataCallback);

  /*
   * close file
   */

  if (havehffile) {
    HTCloseFile ();
    havehffile = 0;
  }

  /*
   * destroy table
   */

  if (lookuptable) {
    for (index=0; index<tableused; index++)
      if (lookuptable[index].used == 1)
	free (lookuptable[index].domnam);
    
    free (lookuptable);

    lookuptable = NULL;
    tableused   = 0;
  }
}

static int
LookupCompare (const void *e1, const void *e2)
{
  if (((struct lookuptable_ *) e1)->address < 
      ((struct lookuptable_ *) e2)->address)
    return -1;
  else if (((struct lookuptable_ *) e1)->address >
	   ((struct lookuptable_ *) e2)->address)
    return 1;

  return 0;
}

/*
 * look up address, first in internal table, then in file, then do the
 * hard lookup
 */

char *
LookupDomain (char *byteaddr)
{
  struct lookuptable_ *result, dummy;
  int deleted, index, cutout, ind2;
  char *thename, buffer[64];
  unsigned int ade[4];
  unsigned long addr;
  struct hostent *he;
  time_t now, deadline;

  if (byteaddr == NULL)
    return NULL;
  if (sscanf (byteaddr, "%u.%u.%u.%u", 
	      &(ade[0]), &(ade[1]), &(ade[2]), &(ade[3])) != 4)
    return byteaddr;

#ifdef DEBUG2
  totlookup++;
#endif

#ifdef DEBUG2
  fprintf (stderr, "LookupDomain (%u.%u.%u.%u)\n",
	   ade[0], ade[1], ade[2], ade[3]);
#endif

  dummy.address = addr = (ade[0]<<24) | (ade[1]<<16) | (ade[2]<<8) | ade[3];

  /*
   * try to look up in our lookup table
   */

  if ((result = bsearch (&dummy, lookuptable, 
			 tableused, sizeof (struct lookuptable_),
			 LookupCompare)) != NULL) {
#ifdef DEBUG2
    fprintf (stderr, "LookupDomain () found in table %s\n",
	     result->domnam ? result->domnam : "(unresolved)");
#endif
    result->used += 1;

#ifdef DEBUG2
    totintable++;
#endif

    if (result->domnam == NULL) {
      return byteaddr;
    }
    else {
      return result->domnam;
    }
  }

  /*
   * try to look up in the hosts file
   */

  if (havehffile) {
    if ((thename = HTLookupEntry (addr, buffer)) != NULL) {
#ifdef DEBUG2
      fprintf (stderr, "LookupDomain () found in hosts file %s\n",
	       thename);
#endif
#ifdef DEBUG2
      totinhfile++;
#endif
      return thename;
    }
  }

  /*
   * unknown ip address. if we don't want DNS lookups, stop here
   */

  if (!hardlookup)
    return NULL;

  /*
   * check first if the table is full. if it is, clean it up
   */

  now = time (NULL);

  if (tableused == LOOKUPSIZE) {
#ifdef DEBUG
    fprintf (stderr, "LookupDomain () table full - cleaning up\n");
#endif

    /*
     * If we have a hosts file, we simply dump the full table into the
     * file and delete the whole table at once. Else delete only the
     * older recent and lesser-used entries. We repeat this until the
     * table is pretty empty.
     */

    if (havehffile && !hfreadonly) {
      HTAddEntries (AddDataCallback);

      tableused = 0;
      tabletime = now;
    }
    else {
      for (cutout=2; 4*tableused > LOOKUPSIZE && cutout<255; cutout++) {
	deleted  = 0;
	deadline = tabletime + (3 * (now - tabletime)) / 4;

	for (index=0; index<tableused; index++) {
	  if (lookuptable[index].used   > 0 &&
	      lookuptable[index].tstamp < deadline &&
	      lookuptable[index].used   < cutout) {
	    lookuptable[index].used = 0;
	    free (lookuptable[index].domnam);
	    deleted++;
	  }
	}

	/*
	 * compress table
	 */

	if (deleted) {
	  for (index=0, ind2=0; ind2<tableused; ind2++) {
	    if (lookuptable[ind2].used > 0) {
	      if (index != ind2) {
		memcpy (&(lookuptable[index]),
			&(lookuptable[ind2]),
			sizeof (struct lookuptable_));
	      }
	      index++;
	    }
	  }
	  tableused = index;
	}
#ifdef DEBUG2
	fprintf (stderr, "LookupTable () cutout %d deleted %d new size %d\n",
		 cutout, deleted, tableused);
#endif
      }
    }
#ifdef DEBUG2
    fprintf (stderr, "LookupTable () cleanup complete\n");
#endif
  }

  /*
   * now do the DNS lookup
   */

#ifdef DEBUG2
  totdnslu++;
#endif

  if ((he = gethostbyaddr ((char *) &addr, 4, AF_INET)) == NULL) {
#ifdef DEBUG2
    fprintf (stderr, "LookupTable () couldn't resolve IP %u.%u.%u.%u\n",
	     ade[0], ade[1], ade[2], ade[3]);
#endif
  }
  else {
#ifdef DEBUG2
    totdnssucc++;
#endif
  }

  /*
   * successfully resolved host name (or not). add to table
   */
  
  if (tableused < LOOKUPSIZE) {
    /*
     * we now must bsearch() ourselves. need the first item larger than
     * the new address. index is lower bound, ind2 is upper bound, cutout
     * is pivot. we know that the element itself is not in the list
     */
    index  = 0;
    ind2   = tableused;
    cutout = (ind2 + index) / 2;

    while (ind2 - index > 1) {
      if (addr < lookuptable[cutout].address) {
	ind2   = cutout;
	cutout = (ind2 + index) / 2;
      }
      else {
	index  = cutout;
	cutout = (ind2 + index) / 2;
      }
    }

    /*
     * if cutout points to an element smaller than addr, the element
     * following cutout is larger than addr
     */

    if (cutout < tableused && addr > lookuptable[cutout].address)
      cutout++;

    /*
     * insert item here
     */
/*
 *    _FP_memmove (&(lookuptable[cutout+1]),
 *		   &(lookuptable[cutout]),
 *		   (tableused - cutout) * sizeof (struct lookuptable_));
 */

    for (index=tableused; index>cutout; index--)
      memcpy (&(lookuptable[index]),
	      &(lookuptable[index-1]),
	      sizeof (struct lookuptable_));

    lookuptable[cutout].used    = 1;
    lookuptable[cutout].tstamp  = now;
    lookuptable[cutout].address = addr;
    lookuptable[cutout].domnam  = (he) ? _FP_strdup (he->h_name) : NULL;
    tableused += 1;

#ifdef DEBUG2
    fprintf (stderr, "LookupEntry () New entry %u.%u.%u.%u = %s at %d\n",
	     ade[0], ade[1], ade[2], ade[3],
	     (he) ? he->h_name : "(unknown)",
	     cutout);
#endif

    /*
     * sanity check
     */
#ifdef DEBUG2
    for (index=0; index<tableused-1; index++) {
      if (lookuptable[index].address >= lookuptable[index+1].address) {
	fprintf (stderr, "LookupEntry () Sanity Check failed at %d %lu.%lu.%lu.%lu %lu.%lu.%lu.%lu\n",
		 index,
		 (lookuptable[index].address>>24)&255,
		 (lookuptable[index].address>>16)&255,
		 (lookuptable[index].address>> 8)&255,
		 (lookuptable[index].address    )&255,
		 (lookuptable[index+1].address>>24)&255,
		 (lookuptable[index+1].address>>16)&255,
		 (lookuptable[index+1].address>> 8)&255,
		 (lookuptable[index+1].address    )&255);
	exit(1);
      }
    }
#endif
  }

  /*
   * done here
   */

  return (he) ? he->h_name : NULL;
}

/*
 * end of lookup functions
 */

#else /* ifdef RESOLVEADDR */

#include <stdio.h>

/*
 * if no address resolution is wanted, define dummy functions.
 */

void InitLookup (char *hfile, int ro, int logip, int dolookup)
{}

void DestroyLookup (void)
{}

char *LookupDomain (char *byteaddr)
{ return NULL; }

#endif
